import json

from jet_bridge_base.encoders import JSONEncoder
from jet_bridge_base.utils.classes import issubclass_safe
from sqlalchemy import Column, text, ForeignKey
from sqlalchemy.exc import DBAPIError, SQLAlchemyError
from sqlalchemy.sql.ddl import AddConstraint, DropConstraint
from sqlalchemy.sql import sqltypes

from jet_bridge_base import status
from jet_bridge_base.db import get_mapped_base, get_engine, reload_request_mapped_base
from jet_bridge_base.exceptions.not_found import NotFound
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.serializers.table import TableColumnSerializer
from jet_bridge_base.utils.db_types import map_to_sql_type, db_to_sql_type, get_sql_type_convert
from jet_bridge_base.views.base.api import APIView
from jet_bridge_base.views.model_description import map_table_column


def map_dto_column(table_name, column, metadata):
    args = []
    column_kwargs = {}
    autoincrement = False
    server_default = None
    db_type = column.get('db_field')
    map_type = column.get('field')
    column_type = db_to_sql_type(db_type) if db_type else map_to_sql_type(map_type)

    if column.get('primary_key', False):
        autoincrement = True

    if 'default_type' in column:
        if column['default_type'] == 'value':
            server_default = column['default_value']

            if isinstance(server_default, bool):
                server_default = '1' if server_default else '0'
        elif column['default_type'] == 'datetime_now':
            server_default = text('NOW()')
        elif column['default_type'] == 'uuid':
            server_default = text("uuid_generate_v4()")
        elif column['default_type'] == 'sequence':
            server_default = text("nextval({})".format(column['default_value']))
        elif column['default_type'] == 'auto_increment':
            autoincrement = True

    params = column.get('params')
    if params:
        if 'length' in params:
            column_kwargs['length'] = params['length']

    try:
        from geoalchemy2 import types
        if column_type is types.Geography:
            column_kwargs['geometry_type'] = 'POINT'
    except ImportError:
        pass

    if issubclass_safe(column_type, sqltypes.DateTime):
        column_kwargs['timezone'] = True

    if callable(column_type):
        try:
            column_type = column_type(**column_kwargs)
        except TypeError:
            pass

    if params:
        if 'related_model' in params:
            model = params['related_model'].get('model')

            try:
                table = list(filter(lambda x: x.name == model, metadata.tables.values()))[0]

                table_primary_keys = table.primary_key.columns.keys()
                table_primary_key = table_primary_keys[0] if len(table_primary_keys) > 0 else None
                related_column_name = params.get('custom_primary_key') or table_primary_key
                related_column = [x for x in table.columns if x.name == related_column_name][0]
                name = '{}_{}_fkey'.format(table_name, column['name'])

                column_type = related_column.type
                foreign_key = ForeignKey(related_column, name=name)
                args.append(foreign_key)
            except IndexError:
                pass

    data_source_meta = {}
    data_source_field = column.get('data_source_field')
    data_source_name = column.get('data_source_name')
    data_source_params = column.get('data_source_params')

    if data_source_field:
        data_source_meta['field'] = data_source_field

    if data_source_name:
        data_source_meta['name'] = data_source_name

    if data_source_params:
        data_source_meta['params'] = data_source_params

    if len(data_source_meta.keys()):
        comment = json.dumps(data_source_meta, cls=JSONEncoder)
    else:
        comment = None

    return Column(
        *args,
        name=column['name'],
        type_=column_type,
        autoincrement=autoincrement,
        primary_key=column.get('primary_key', False),
        nullable=column.get('null', False),
        server_default=server_default,
        comment=comment
    )


class TableColumnView(APIView):
    permission_classes = (HasProjectPermissions,)

    def get_db(self, request):
        MappedBase = get_mapped_base(request)
        engine = get_engine(request)
        return MappedBase.metadata, engine

    def update_base(self, request):
        reload_request_mapped_base(request)

    def get_table(self, request):
        metadata, engine = self.get_db(request)
        table_name = request.path_kwargs['table']

        try:
            obj = list(filter(lambda x: x.name == table_name, metadata.tables.values()))[0]
        except IndexError:
            raise NotFound

        self.check_object_permissions(request, obj)

        return obj

    def get_object(self, request):
        metadata, engine = self.get_db(request)
        table_name = request.path_kwargs['table']

        try:
            table = list(filter(lambda x: x.name == table_name, metadata.tables.values()))[0]
        except IndexError:
            raise NotFound

        pk = request.path_kwargs['pk']
        obj = table.columns.get(pk)

        if obj is None:
            raise NotFound

        self.check_object_permissions(request, obj)

        return obj

    def list(self, request, *args, **kwargs):
        metadata, engine = self.get_db(request)
        table = self.get_table(request)
        columns = list(map(lambda x: map_table_column(metadata, table, x, True), table.columns))
        return JSONResponse(columns)

    def retrieve(self, request, *args, **kwargs):
        metadata, engine = self.get_db(request)
        instance = self.get_object(request)
        table = instance.table
        return JSONResponse(map_table_column(metadata, table, instance, True))

    def create(self, request, *args, **kwargs):
        serializer = TableColumnSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            self.perform_create(request, serializer)
        except Exception as e:
            if isinstance(e, DBAPIError):
                if '(psycopg2.errors.DuplicateColumn)' in e.args[0]:
                    raise ValidationError('Column with such name already exists')
            raise ValidationError(str(e))

        return JSONResponse(serializer.representation_data, status=status.HTTP_201_CREATED)

    def perform_create(self, request, serializer):
        metadata, engine = self.get_db(request)
        table = self.get_table(request)
        column = map_dto_column(table.name, serializer.validated_data, metadata)
        column._set_parent(table)

        ddl_compiler = engine.dialect.ddl_compiler(engine.dialect, None)
        column_specification = ddl_compiler.get_column_specification(column)

        table_name = ddl_compiler.preparer.format_table(table)
        engine.execute('''ALTER TABLE {0} ADD COLUMN {1}'''.format(table_name, column_specification))

        for foreign_key in column.foreign_keys:
            if not foreign_key.constraint:
                foreign_key._set_table(column, table)
                engine.execute(AddConstraint(foreign_key.constraint))

        self.update_base(request)

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object(request)
        self.perform_destroy(request, instance)
        return JSONResponse(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, request, column):
        metadata, engine = self.get_db(request)
        table = self.get_table(request)

        ddl_compiler = engine.dialect.ddl_compiler(engine.dialect, None)
        table_name = ddl_compiler.preparer.format_table(table)
        column_name = ddl_compiler.preparer.format_column(column)
        engine.execute('''ALTER TABLE {0} DROP COLUMN {1}'''.format(table_name, column_name))

        table._columns.remove(column)

        for fk in column.foreign_keys:
            table.foreign_keys.remove(fk)
            if fk.constraint in table.constraints:
                # this might have been removed
                # already, if it's a composite constraint
                # and more than one col being replaced
                table.constraints.remove(fk.constraint)

        self.update_base(request)

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object(request)
        serializer = TableColumnSerializer(data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)

        try:
            self.perform_update(request, instance, serializer)
        except Exception as e:
            if isinstance(e, DBAPIError):
                if '(psycopg2.errors.InvalidDatetimeFormat)' in e.args[0]:
                    raise ValidationError('Some of the rows has invalid date format')
                elif '(psycopg2.errors.DuplicateColumn)' in e.args[0]:
                    raise ValidationError('Column with such name already exists')
            raise ValidationError(str(e))

        return JSONResponse(serializer.representation_data)

    def perform_update(self, request, existing_column, serializer):
        metadata, engine = self.get_db(request)
        table = self.get_table(request)
        existing_data = map_table_column(metadata, table, existing_column, True)
        existing_dto = {
            'name': existing_data['name'],
            'field': existing_data['field'],
            'primary_key': existing_column.table.primary_key.columns[0].name == existing_data['name']
        }

        if 'length' in existing_data:
            existing_dto['length'] = existing_data['length']

        column_data = {
            **existing_dto,
            **serializer.validated_data
        }
        column = map_dto_column(table.name, column_data, metadata)

        ddl_compiler = engine.dialect.ddl_compiler(engine.dialect, None)
        table_name = ddl_compiler.preparer.format_table(table)

        column_name = ddl_compiler.preparer.format_column(column)
        column_type = column.type.compile(engine.dialect)
        existing_column_name = ddl_compiler.preparer.format_column(existing_column)
        existing_column_type = existing_column.type.compile(engine.dialect)

        preserve_column_index = None

        if column_name != existing_column_name:
            try:
                preserve_column_index = table._columns.values().index(existing_column)
                table._columns.remove(existing_column)
            except ValueError:
                pass

        column._set_parent(table)

        if preserve_column_index is not None:
            try:
                new_column_index = table._columns.values().index(column)
                columns_collection = table._columns._collection
                columns_collection.insert(preserve_column_index, columns_collection.pop(new_column_index))
            except ValueError:
                pass

        column_type_stmt = column_type
        sql_type_convert = get_sql_type_convert(column.type)

        if sql_type_convert:
            column_type_stmt += ' USING {0}'.format(sql_type_convert(existing_column_name))

        with request.session.begin():
            try:
                for foreign_key in existing_column.foreign_keys:
                    if foreign_key.constraint:
                        foreign_key_should_exist = any(map(lambda x: x.target_fullname == foreign_key.target_fullname, column.foreign_keys))
                        if foreign_key_should_exist:
                            continue
                        request.session.execute(DropConstraint(foreign_key.constraint))

                if column_type != existing_column_type:
                    request.session.execute('''ALTER TABLE {0} ALTER COLUMN {1} DROP DEFAULT'''.format(table_name, existing_column_name))
                    request.session.execute('''ALTER TABLE {0} ALTER COLUMN {1} TYPE {2}'''.format(table_name, existing_column_name, column_type_stmt))

                if column.nullable:
                    request.session.execute('''ALTER TABLE {0} ALTER COLUMN {1} DROP NOT NULL'''.format(table_name, existing_column_name))
                else:
                    request.session.execute('''ALTER TABLE {0} ALTER COLUMN {1} SET NOT NULL'''.format(table_name, existing_column_name))

                default = ddl_compiler.get_column_default_string(column)

                if default is not None:
                    request.session.execute('''ALTER TABLE {0} ALTER COLUMN {1} SET DEFAULT {2}'''.format(table_name, existing_column_name, default))
                else:
                    request.session.execute('''ALTER TABLE {0} ALTER COLUMN {1} DROP DEFAULT'''.format(table_name, existing_column_name))

                column_rename = column_name != existing_column_name

                if column_rename:
                    request.session.execute('''ALTER TABLE {0} RENAME COLUMN {1} TO {2}'''.format(table_name, existing_column_name, column_name))

                for foreign_key in column.foreign_keys:
                    if not foreign_key.constraint:
                        foreign_key_exists = any(map(lambda x: x.target_fullname == foreign_key.target_fullname, existing_column.foreign_keys))
                        if foreign_key_exists:
                            continue
                        foreign_key._set_table(column, table)
                        request.session.execute(AddConstraint(foreign_key.constraint))

                self.update_base(request)
            except SQLAlchemyError as e:
                request.session.rollback()

                table._columns.remove(column)
                existing_column._set_parent(table)

                for fk in column.foreign_keys:
                    table.foreign_keys.remove(fk)
                    if fk.constraint in table.constraints:
                        table.constraints.remove(fk.constraint)

                self.update_base(request)
                raise e

    def partial_update(self, *args, **kwargs):
        kwargs['partial'] = True
        return self.update(*args, **kwargs)



import json
import re
import time

import requests
from social_core.backends.base import BaseAuth
from social_core.utils import module_member

from jet_bridge_base import fields, settings
from jet_bridge_base.configuration import configuration
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.external_auth.utils import load_strategy
from jet_bridge_base.responses.base import Response
from jet_bridge_base.serializers.serializer import Serializer
from jet_bridge_base.utils.backend import get_secret_tokens
from jet_bridge_base.utils.compress import compress_data, decompress_data


class ProxyRequestSerializer(Serializer):
    method = fields.CharField(required=False, default='GET')
    url = fields.CharField()
    query_params = fields.JSONField(required=False, default=dict)
    headers = fields.JSONField(required=False, default=dict)
    body = fields.RawField(required=False)

    project = fields.CharField(required=False)
    environment = fields.CharField(required=False)
    resource = fields.CharField(required=False)
    draft = fields.BooleanField(required=False, default=False)
    secret_tokens = fields.CharField(required=False)
    context = fields.JSONField(required=False)

    def get_access_token(self, app, config, extra_data):
        handler = self.context.get('handler')
        request = self.context.get('request')
        strategy = load_strategy(handler, request, config)

        backend_path = config.get('backend_path')
        Backend = module_member(backend_path)

        if not issubclass(Backend, BaseAuth):
            return extra_data.get('access_token')

        backend = Backend(strategy, None)

        refresh_token = extra_data.get('refresh_token') or extra_data.get('access_token')

        expires_on = None
        params_expires_on = extra_data.get('expires_on')
        params_token_updated = extra_data.get('token_updated')
        params_expires_in = extra_data.get('expires') or extra_data.get('expires_in')

        try:
            if params_expires_on:
                expires_on = int(params_expires_on)
            elif params_expires_in and params_token_updated:
                expires_on = int(params_token_updated) + int(params_expires_in)
        except (ValueError, TypeError):
            pass

        try:
            if refresh_token and (not expires_on or expires_on <= int(time.time())):
                response = backend.refresh_token(token=refresh_token)
                if not backend.EXTRA_DATA or len(backend.EXTRA_DATA) == 0:
                    backend.GET_ALL_EXTRA_DATA = True
                new_extra_data = backend.extra_data(user=None, uid=None, response=response, details={})
                access_token = new_extra_data.get('access_token')

                new_extra_data = {
                    'expires_on': new_extra_data.get('expires_on'),
                    'access_token': new_extra_data.get('access_token'),
                    'expires': new_extra_data.get('expires'),
                    'auth_time': new_extra_data.get('auth_time'),
                    'refresh_token': new_extra_data.get('refresh_token'),
                    'token_updated': int(time.time())
                }

                request = self.context.get('request')
                extra_data_key = '_'.join(['extra_data', app])

                new_extra_data_str = json.dumps(new_extra_data)

                if settings.COOKIE_COMPRESS:
                    new_extra_data_str = compress_data(new_extra_data_str)

                configuration.session_set(request, extra_data_key, new_extra_data_str, secure=not settings.COOKIE_COMPRESS)

                return access_token
        except Exception:
            pass

        return extra_data.get('access_token')

    def resolve_secret_tokens(self, names, project_name, environment_name, resource, draft):
        request = self.context.get('request')
        instances = {}
        unresolved = names[:]

        for name in unresolved:
            regex = re.search('sso\.(?P<sso>\w+)\.(?P<token>\w+)', name)

            if not regex:
                continue

            matches = regex.groupdict()

            if matches['token'] != 'access_token':
                continue

            app = configuration.clean_sso_application_name(matches['sso'])
            config = settings.SSO_APPLICATIONS.get(app)

            if not config:
                continue

            extra_data_key = '_'.join(['extra_data', app])

            try:
                if settings.COOKIE_COMPRESS:
                    extra_data_str = configuration.session_get(request, extra_data_key, decode=False, secure=False)
                    extra_data_str = decompress_data(extra_data_str)
                else:
                    extra_data_str = configuration.session_get(request, extra_data_key)

                extra_data = json.loads(extra_data_str)

                if matches['token'] not in extra_data:
                    continue

                if matches['token'] == 'access_token':
                    instances[name] = self.get_access_token(app, config, extra_data)
                else:
                    instances[name] = extra_data.get(matches['token'])
            except Exception:
                pass

        unresolved = list(filter(lambda x: x not in instances.keys(), unresolved))

        if len(unresolved):
            token_prefix = 'Token '
            authorization = request.headers.get('AUTHORIZATION', '')
            user_token = authorization[len(token_prefix):] if authorization.startswith(token_prefix) else None

            for item in get_secret_tokens(project_name, environment_name, resource, draft, settings.TOKEN, user_token):
                if item['name'] not in unresolved:
                    continue
                instances[item['name']] = item['value']

        for name in unresolved:
            if name not in instances:
                instances[name] = ''

        return instances

    def validate(self, attrs):
        if 'resource' in attrs:
            if 'project' not in attrs:
                raise ValidationError('"project" is required when specifying "resource"')

        if 'secret_tokens' in attrs and len(attrs['secret_tokens']):
            if 'resource' not in attrs:
                raise ValidationError('"resource" is required when specifying "secret_tokens"')
            names = attrs['secret_tokens'].split(',')

            attrs['secret_tokens'] = self.resolve_secret_tokens(
                names,
                attrs.get('project'),
                attrs.get('environment'),
                attrs.get('resource'),
                attrs.get('draft')
            )

        if isinstance(attrs['headers'], dict):
            attrs['headers'] = dict([[key, str(value)] for key, value in attrs['headers'].items()])

        return attrs

    def params_dict_to_list(self, items):
        return list(map(lambda x: {'name': x[0], 'value': x[1]}, items.items()))

    def interpolate(self, url, headers, query_params, body, pattern, replaces):
        def replace(str, replaces):
            for name, value in replaces.items():
                str = str.replace(pattern % name, value)
            return str

        url = replace(url, replaces)

        for header in headers:
            header['value'] = replace(header['value'], replaces)

        for query_param in query_params:
            query_param['value'] = replace(query_param['value'], replaces)

        if body:
            body = replace(body, replaces)

        return url, headers, query_params, body

    def submit(self):
        method = self.validated_data['method']
        url = self.validated_data['url']
        headers = self.validated_data['headers'] or []
        query_params = self.validated_data['query_params'] or []
        body = self.validated_data.get('body')

        if isinstance(headers, dict):
            headers = self.params_dict_to_list(headers)

        if isinstance(query_params, dict):
            query_params = self.params_dict_to_list(query_params)

        if 'secret_tokens' in self.validated_data:
            url, headers, query_params, body = self.interpolate(url, headers, query_params, body, '{-%s-}', self.validated_data['secret_tokens'])

        if 'context' in self.validated_data:
            url, headers, query_params, body = self.interpolate(url, headers, query_params, body, '{{%s}}', self.validated_data['context'])

        if body:
            body = body.encode('utf-8')

        headers = dict(map(lambda x: (x['name'], x['value']), headers))
        query_params = list(map(lambda x: (x['name'], x['value']), query_params))

        try:
            r = requests.request(method, url, headers=headers, params=query_params, data=body)
            response_headers = r.headers

            remove_headers = [
                'Access-Control-Allow-Origin',
                'Access-Control-Allow-Methods',
                'Access-Control-Allow-Headers',
                'Access-Control-Expose-Headers',
                'Access-Control-Allow-Credentials',
                'Connection',
                'Content-Encoding',
                'Content-Length',
                'Keep-Alive',
                'Proxy-Authenticate',
                'Proxy-Authorization',
                'TE',
                'Trailers',
                'Transfer-Encoding',
                'Upgrade'
            ]

            for header in remove_headers:
                if header in response_headers:
                    del response_headers[header]

            response = Response(data=r.content, status=r.status_code, headers=response_headers)

            return response
        except Exception as e:
            raise ValidationError(e)



import json


class MongoOperator(object):
    def __init__(self, operator, lhs=None, rhs=None):
        self.operator = operator
        self.lhs = lhs
        self.rhs = rhs

    def __invert__(self):
        return MongoOperator('not', self)

    def __repr__(self):
        return json.dumps({
            'operator': self.operator,
            'lhs': repr(self.lhs),
            'rhs': repr(self.rhs),
        })



from social_core.storage import BaseStorage, UserMixin


class User(object):
    social_user = None
    is_new = True
    details = None
    extra_data = None


class UserController(UserMixin):

    @classmethod
    def user_model(cls):
        return User


class JetBridgeStorage(BaseStorage):
    user = UserController



import six

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from jet_bridge_base.exceptions.validation_error import ValidationError


class empty:
    """
    This class is used to represent no data being provided for a given input
    or output value.

    It is required because `None` may be a valid input or output value.
    """
    pass


class Field(object):
    creation_counter = 0
    field_name = None
    field_error_messages = {
        'required': 'this field is required',
        'null': 'this field may not be null'
    }

    def __init__(self, *args, **kwargs):
        self.creation_counter = Field.creation_counter
        Field.creation_counter += 1

        self.required = kwargs.pop('required', True)
        self.read_only = kwargs.pop('read_only', False)
        self.write_only = kwargs.pop('write_only', False)
        self.many = kwargs.pop('many', False)
        self.allow_many = kwargs.pop('allow_many', False)
        self.default = kwargs.pop('default', empty)
        self.context = kwargs.pop('context', {})

        messages = {}
        for cls in reversed(self.__class__.__mro__):
            messages.update(getattr(cls, 'field_error_messages', {}))
        self.error_messages = messages

    def validate(self, value):
        return value

    def get_default(self):
        if callable(self.default):
            return self.default()
        return self.default

    def get_value(self, data):
        try:
            if isinstance(data, Mapping):
                field_value = data[self.field_name]
            else:
                field_value = getattr(data, self.field_name)
        except (KeyError, AttributeError):
            if self.default is not empty:
                return self.get_default()
            else:
                return empty

        if not getattr(self, 'many', False) and not getattr(self, 'allow_many', False) and isinstance(field_value, list):
            field_value = self.get_non_array_value(field_value)

        if isinstance(field_value, bytes):
            field_value = field_value.decode('utf8')

        return field_value

    def get_non_array_value(self, array_value):
        result = []

        for value in array_value:
            if value is None:
                continue
            str_value = six.text_type(value)
            result.append(str_value)

        return ','.join(result)

    def run_validation(self, value):
        if value is empty:
            if self.required:
                # raise ValidationError('Field is required')
                self.error('required')
            else:
                return None
        return self.to_internal_value(value)

    def to_internal_value_item(self, value):
        raise NotImplementedError

    def to_internal_value(self, value):
        if self.many:
            if value is empty:
                return []
            return list(map(lambda x: self.to_internal_value_item(x), value))
        else:
            return self.to_internal_value_item(value)

    def to_representation_item(self, value):
        raise NotImplementedError

    def to_representation(self, value):
        if self.many:
            return list(map(lambda x: self.to_representation_item(x), value or []))
        else:
            return self.to_representation_item(value)

    def error(self, key, **kwargs):
        """
        A helper method that simply raises a validation error.
        """
        try:
            msg = self.error_messages[key]
        except KeyError:
            class_name = self.__class__.__name__
            raise AssertionError('Error with key={} is not found for class={}'.format(key, class_name))
        message_string = msg.format(**kwargs)
        raise ValidationError(message_string, code=key)



CHAR = 'CharField'
FIXED_CHAR = 'FixedCharField'
TEXT = 'TextField'
UUID = 'UUIDField'
SELECT = 'SelectField'
BOOLEAN = 'BooleanField'
INTEGER = 'IntegerField'
BIG_INTEGER = 'BigIntegerField'
SMALL_INTEGER = 'SmallIntegerField'
FLOAT = 'FloatField'
NUMBER = 'NumberField'
DECIMAL = 'DecimalField'
DOUBLE_PRECISION = 'DoublePrecisionField'
MONEY = 'MoneyField'
DATE = 'DateField'
DATE_TIME = 'DateTimeField'
TIMESTAMP = 'TimestampField'
FOREIGN_KEY = 'ForeignKey'
JSON = 'JSONField'
GEOMETRY = 'GeometryField'
GEOGRAPHY = 'GeographyField'
BINARY = 'BinaryField'



from six.moves.urllib_parse import quote

from jet_bridge_base import settings
from jet_bridge_base.responses.base import Response
from jet_bridge_base.responses.redirect import RedirectResponse
from jet_bridge_base.status import HTTP_400_BAD_REQUEST
from jet_bridge_base.views.base.api import BaseAPIView


class RegisterView(BaseAPIView):

    def get(self, request, *args, **kwargs):
        if not settings.PROJECT:
            return Response('Project name is not set', status=HTTP_400_BAD_REQUEST)

        if not settings.TOKEN:
            return Response('Project token is not set', status=HTTP_400_BAD_REQUEST)

        token = request.get_argument('token', '')
        environment_type = settings.ENVIRONMENT_TYPE

        if settings.WEB_BASE_URL.startswith('https') and not request.full_url().startswith('https'):
            web_base_url = 'http{}'.format(settings.WEB_BASE_URL[5:])
        else:
            web_base_url = settings.WEB_BASE_URL

        if settings.ENVIRONMENT:
            url = '{}/builder/{}/{}/resources/database/create/'.format(web_base_url, settings.PROJECT, settings.ENVIRONMENT)
        else:
            url = '{}/builder/{}/resources/database/create/'.format(web_base_url, settings.PROJECT)

        parameters = [
            ['engine', settings.DATABASE_ENGINE],
            ['referrer', request.full_url().encode('utf8')],
        ]

        if token:
            parameters.append(['token', token])

        if environment_type:
            parameters.append(['environment_type', environment_type])

        query_string = '&'.join(map(lambda x: '{}={}'.format(x[0], quote(x[1])), parameters))

        return RedirectResponse('%s#%s' % (url, query_string))



from sqlalchemy.sql import sqltypes

from jet_bridge_base.db_types import MongoColumn
from jet_bridge_base.filters import lookups
from jet_bridge_base.filters.binary_filter import BinaryFilter
from jet_bridge_base.filters.float_filter import FloatFilter
from jet_bridge_base.filters.boolean_filter import BooleanFilter
from jet_bridge_base.filters.char_filter import CharFilter
from jet_bridge_base.filters.datetime_filter import DateTimeFilter
from jet_bridge_base.filters.wkt_filter import WKTFilter
from jet_bridge_base.filters.integer_filter import IntegerFilter
from jet_bridge_base.models import data_types

number_lookups = [
    lookups.EXACT,
    lookups.GT,
    lookups.GTE,
    lookups.LT,
    lookups.LTE,
    lookups.ICONTAINS,
    lookups.IN,
    lookups.IS_NULL,
    lookups.IS_EMPTY,
]

datetime_lookups = [
    lookups.EXACT,
    lookups.GT,
    lookups.GTE,
    lookups.LT,
    lookups.LTE,
    lookups.ICONTAINS,
    lookups.IN,
    lookups.IS_NULL,
    lookups.IS_EMPTY,
]

text_lookups = [
    lookups.EXACT,
    lookups.ICONTAINS,
    lookups.IN,
    lookups.STARTS_WITH,
    lookups.ENDS_WITH,
    lookups.IS_NULL,
    lookups.IS_EMPTY,
]

select_lookups = [
    lookups.EXACT,
    lookups.ICONTAINS,
    lookups.IN,
    lookups.STARTS_WITH,
    lookups.ENDS_WITH,
    lookups.IS_NULL,
    lookups.IS_EMPTY,
]

boolean_lookups = [
    lookups.EXACT,
    lookups.IN,
    lookups.IS_NULL,
    lookups.IS_EMPTY,
]

json_lookups = [
    lookups.EXACT,
    lookups.IN,
    lookups.JSON_ICONTAINS,
    lookups.IS_NULL,
    lookups.IS_EMPTY,
]

geography_lookups = [
    lookups.COVEREDBY
]

binary_lookups = [
    lookups.EXACT,
    lookups.IN,
    lookups.IS_NULL,
    lookups.IS_EMPTY,
]

FILTER_FOR_DBFIELD = {
    sqltypes.VARCHAR: {'filter_class': CharFilter, 'lookups': text_lookups, 'lookups_name': 'text'},
    sqltypes.TEXT: {'filter_class': CharFilter, 'lookups': text_lookups, 'lookups_name': 'text'},
    sqltypes.Enum: {'filter_class': CharFilter, 'lookups': select_lookups, 'lookups_name': 'select'},
    sqltypes.Boolean: {'filter_class': BooleanFilter, 'lookups': boolean_lookups, 'lookups_name': 'boolean'},
    sqltypes.Integer: {'filter_class': IntegerFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    sqltypes.SmallInteger: {'filter_class': IntegerFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    sqltypes.BigInteger: {'filter_class': IntegerFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    sqltypes.Numeric: {'filter_class': IntegerFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    sqltypes.Float: {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    sqltypes.Date: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'},
    sqltypes.DateTime: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'},
    sqltypes.TIMESTAMP: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'},
    sqltypes.JSON: {'filter_class': CharFilter, 'lookups': json_lookups, 'lookups_name': 'json'},
    sqltypes.ARRAY: {'filter_class': CharFilter, 'lookups': json_lookups, 'lookups_name': 'json'},
    sqltypes.BINARY: {'filter_class': BinaryFilter, 'lookups': binary_lookups, 'lookups_name': 'binary'},
    # sqlalchemy.TextField:                   {'filter_class': CharFilter},
    # sqlalchemy.BOOLEAN:                {'filter_class': BooleanFilter},
    # sqlalchemy.DateField:                   {'filter_class': DateFilter},
    # sqlalchemy.DateTimeField:               {'filter_class': DateTimeFilter},
    # sqlalchemy.TimeField:                   {'filter_class': TimeFilter},
    # sqlalchemy.DurationField:               {'filter_class': DurationFilter},
    # sqlalchemy.DecimalField:                {'filter_class': NumberFilter},
    # sqlalchemy.SmallIntegerField:           {'filter_class': NumberFilter},
    # sqlalchemy.IntegerField:                {'filter_class': NumberFilter},
    # sqlalchemy.PositiveIntegerField:        {'filter_class': NumberFilter},
    # sqlalchemy.PositiveSmallIntegerField:   {'filter_class': NumberFilter},
    # sqlalchemy.FloatField:                  {'filter_class': NumberFilter},
    # sqlalchemy.NullBooleanField:            {'filter_class': BooleanFilter},
    # sqlalchemy.SlugField:                   {'filter_class': CharFilter},
    # sqlalchemy.EmailField:                  {'filter_class': CharFilter},
    # sqlalchemy.FilePathField:               {'filter_class': CharFilter},
    # sqlalchemy.URLField:                    {'filter_class': CharFilter},
    # sqlalchemy.GenericIPAddressField:       {'filter_class': CharFilter},
    # sqlalchemy.CommaSeparatedIntegerField:  {'filter_class': CharFilter},
    # sqlalchemy.UUIDField:                   {'filter_class': UUIDFilter}
}
FILTER_FOR_DBFIELD_DEFAULT = FILTER_FOR_DBFIELD[sqltypes.VARCHAR]

try:
    from geoalchemy2 import types
    FILTER_FOR_DBFIELD[types.Geometry] = {'filter_class': WKTFilter, 'lookups': geography_lookups, 'lookups_name': 'geography'}
    FILTER_FOR_DBFIELD[types.Geography] = {'filter_class': WKTFilter, 'lookups': geography_lookups, 'lookups_name': 'geography'}
except ImportError:
    pass

try:
    from sqlalchemy.dialects import mssql
    FILTER_FOR_DBFIELD[mssql.MONEY] = {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'}
    FILTER_FOR_DBFIELD[mssql.SMALLMONEY] = {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'}
except ImportError:
    pass

try:
    from databricks.sqlalchemy.dialect import DatabricksDecimal, DatabricksDate, DatabricksTimestamp

    FILTER_FOR_DBFIELD[DatabricksDecimal] = {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'}
    FILTER_FOR_DBFIELD[DatabricksDate] = {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'}
    FILTER_FOR_DBFIELD[DatabricksTimestamp] = {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'}
except ImportError:
    pass


FILTER_FOR_MAP_TYPE = {
    data_types.CHAR: {'filter_class': CharFilter, 'lookups': text_lookups, 'lookups_name': 'text'},
    data_types.FIXED_CHAR: {'filter_class': CharFilter, 'lookups': text_lookups, 'lookups_name': 'text'},
    data_types.TEXT: {'filter_class': CharFilter, 'lookups': text_lookups, 'lookups_name': 'text'},
    data_types.UUID: {'filter_class': BinaryFilter, 'lookups': binary_lookups, 'lookups_name': 'binary'},
    data_types.SELECT: {'filter_class': CharFilter, 'lookups': select_lookups, 'lookups_name': 'select'},
    data_types.BOOLEAN: {'filter_class': BooleanFilter, 'lookups': boolean_lookups, 'lookups_name': 'boolean'},
    data_types.INTEGER: {'filter_class': IntegerFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.BIG_INTEGER: {'filter_class': IntegerFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.SMALL_INTEGER: {'filter_class': IntegerFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.FLOAT: {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.NUMBER: {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.DECIMAL: {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.DOUBLE_PRECISION: {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.MONEY: {'filter_class': FloatFilter, 'lookups': number_lookups, 'lookups_name': 'number'},
    data_types.DATE: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'},
    data_types.DATE_TIME: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'},
    data_types.TIMESTAMP: {'filter_class': DateTimeFilter, 'lookups': datetime_lookups, 'lookups_name': 'datetime'},
    data_types.FOREIGN_KEY: {'filter_class': CharFilter, 'lookups': text_lookups, 'lookups_name': 'text'},
    data_types.JSON: {'filter_class': CharFilter, 'lookups': json_lookups, 'lookups_name': 'json'},
    data_types.GEOMETRY: {'filter_class': WKTFilter, 'lookups': geography_lookups, 'lookups_name': 'geography'},
    data_types.GEOGRAPHY: {'filter_class': WKTFilter, 'lookups': geography_lookups, 'lookups_name': 'geography'},
    data_types.BINARY: {'filter_class': BinaryFilter, 'lookups': binary_lookups, 'lookups_name': 'binary'},
}

FILTER_FOR_MAP_TYPE_DEFAULT = FILTER_FOR_MAP_TYPE[data_types.CHAR]


def filter_for_data_type(value):
    for data_type, filter_data in FILTER_FOR_DBFIELD.items():
        if isinstance(value, data_type):
            return filter_data
    return FILTER_FOR_DBFIELD_DEFAULT


def filter_for_column(column):
    if isinstance(column, MongoColumn):
        return FILTER_FOR_MAP_TYPE.get(column.type, FILTER_FOR_MAP_TYPE_DEFAULT)
    else:
        return filter_for_data_type(column.type)



from .mongo_operator import MongoOperator


class MongoColumn(object):
    def __init__(self, table, name, type, nullable=True, mixed_types=None, autoincrement=False, default=None,
                 server_default=None, foreign_keys=None, comment=None, params=None):
        self.table = table
        self.name = name
        self.key = name
        self.type = type
        self.nullable = nullable
        self.mixed_types = mixed_types
        self.autoincrement = autoincrement
        self.default = default
        self.server_default = server_default
        self.foreign_keys = foreign_keys or list()
        self.comment = comment
        self.params = params or dict()

    @staticmethod
    def deserialize(table, obj):
        name = obj.pop('name')
        type = obj.pop('type')
        return MongoColumn(table, name, type, **obj)

    def serialize(self):
        return {
            'name': self.name,
            'type': self.type,
            'nullable': self.nullable,
            'mixed_types': self.mixed_types,
            'autoincrement': self.autoincrement,
            'default': self.default,
            'server_default': self.server_default,
            'foreign_keys': self.foreign_keys,
            'comment': self.comment,
            'params': self.params
        }

    def __eq__(self, other):
        return MongoOperator('__eq__', self, other)

    def __gt__(self, other):
        return MongoOperator('__gt__', self, other)

    def __ge__(self, other):
        return MongoOperator('__ge__', self, other)

    def __lt__(self, other):
        return MongoOperator('__lt__', self, other)

    def __le__(self, other):
        return MongoOperator('__le__', self, other)

    def isnot(self, other):
        return MongoOperator('not', MongoOperator('__eq__', self, other))

    def ilike(self, value):
        return MongoOperator('ilike', self, value)

    def exists(self, value):
        return MongoOperator('exists', self, value)

    def in_(self, value):
        return MongoOperator('in', self, value)

    def json_icontains(self, value):
        return MongoOperator('json_icontains', self, value)

    def __repr__(self):
        return '{}.{}'.format(self.table.name, self.name)



from .mongo_table import MongoTable


class MongoMetadata(object):
    def __init__(self, tables=None, schema=None):
        self.tables = list(map(lambda x: MongoTable.deserialize(x), tables)) if tables else list()
        self.schema = schema

    def append_table(self, table):
        self.tables.append(table)

    @staticmethod
    def deserialize(obj):
        return MongoMetadata(**obj)

    def serialize(self):
        return {
            'tables': list(map(lambda x: x.serialize(), self.tables)),
            'schema': self.schema
        }



from jet_bridge_base.utils.common import CollectionDict


class MongoMapper(object):
    def __init__(self, table):
        self.tables = [table]
        self.relationships = CollectionDict()
        self.selectable = table
        self.primary_key = [table.columns['_id']]
        self.columns = table.columns

    def __repr__(self):
        return self.selectable.name



import base64
import time
from datetime import datetime

from six.moves.urllib_parse import parse_qsl, quote_plus
from sqlalchemy import MetaData, create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from jet_bridge_base.automap import automap_base
from jet_bridge_base.utils.conf import get_connection_only_predicate
from jet_bridge_base.utils.process import get_memory_usage_human, get_memory_usage
from jet_bridge_base.utils.tables import get_table_name
from jet_bridge_base.logger import logger

from .sql_metadata_file import sql_load_metadata_file, sql_dump_metadata_file
from .sql_reflect import sql_reflect
from .timezones import sql_fetch_default_timezone
from .type_codes import fetch_type_code_to_sql_type


def sql_init_database_connection(conf, tunnel, id_short, connection_name, schema, pending_connection):
    engine = sql_create_connection_engine(conf, tunnel)
    pending_connection['engine'] = engine

    Session = scoped_session(sessionmaker(bind=engine))
    session = Session()

    logger.info('[{}] Connecting to database "{}"...'.format(id_short, connection_name))

    connect_start = time.time()
    with session.connection() as connection:
        connect_end = time.time()
        connect_time = round(connect_end - connect_start, 3)

        logger.info('[{}] Getting db types for "{}"...'.format(id_short, connection_name))
        type_code_to_sql_type = fetch_type_code_to_sql_type(session)

        default_timezone = sql_fetch_default_timezone(session)
        default_timezone_updated = datetime.now()

        if default_timezone is not None:
            logger.info('[{}] Default timezone detected: "{}"'.format(id_short, default_timezone))
        else:
            logger.info('[{}] Failed to detect default timezone'.format(id_short))

        metadata_dump = sql_load_metadata_file(conf, connection)

        if metadata_dump:
            metadata = metadata_dump['metadata']

            reflect_time = None
            reflect_memory_usage_approx = None

            logger.info('[{}] Loaded schema cache for "{}"'.format(id_short, connection_name))
        else:
            logger.info('[{}] Getting schema for "{}"...'.format(id_short, connection_name))

            reflect_start_time = time.time()
            reflect_start_memory_usage = get_memory_usage()

            metadata = MetaData(schema=schema, bind=connection)
            only = get_connection_only_predicate(conf)
            sql_reflect(id_short, metadata, engine, only=only, pending_connection=pending_connection, foreign=True, views=True)

            reflect_end_time = time.time()
            reflect_end_memory_usage = get_memory_usage()
            reflect_time = round(reflect_end_time - reflect_start_time, 3)
            reflect_memory_usage_approx = reflect_end_memory_usage - reflect_start_memory_usage

            sql_dump_metadata_file(conf, metadata)

        logger.info('[{}] Connected to "{}" (Mem:{})'.format(id_short, connection_name, get_memory_usage_human()))

        MappedBase = automap_base(metadata=metadata)
        sql_load_mapped_base(MappedBase)

        for table_name, table in MappedBase.metadata.tables.items():
            if len(table.primary_key.columns) == 0 and table_name not in MappedBase.classes:
                logger.warning(
                    '[{}] Table "{}" does not have primary key and will be ignored'.format(id_short, table_name))

        result = {
            'engine': engine,
            'Session': Session,
            'MappedBase': MappedBase,
            'type_code_to_sql_type': type_code_to_sql_type,
            'default_timezone': default_timezone,
            'default_timezone_updated': default_timezone_updated,
            'connect_time': connect_time,
            'reflect_time': reflect_time,
            'reflect_memory_usage_approx': reflect_memory_usage_approx,
            'reflect_metadata_dump': metadata_dump['file_path'] if metadata_dump else None
        }

    session.close()

    return result


def sql_build_engine_url(conf, tunnel=None):
    if not conf.get('engine') or not conf.get('name'):
        return

    url = [
        str(conf.get('engine')),
        '://'
    ]

    if conf.get('engine') == 'sqlite':
        url.append('/')
        url.append(str(conf.get('name')))

        if conf.get('extra'):
            url.append('?')
            url.append(str(conf.get('extra')))
    elif conf.get('engine') == 'bigquery':
        url.append(str(conf.get('name')))

        try:
            base64.b64decode(conf.get('password'))
            url.append('?credentials_base64={}'.format(conf.get('password')))

            if conf.get('extra'):
                url.append('&')
                url.append(str(conf.get('extra')))
        except:
            pass
    elif conf.get('engine') == 'snowflake':
        url.append(url_encode(str(conf.get('user'))))

        if conf.get('password'):
            url.append(':')
            url.append(url_encode(str(conf.get('password'))))

        url.append('@')

        url.append(str(conf.get('host')))
        url.append('/')

        url.append(str(conf.get('name')))

        if conf.get('extra'):
            url.append('?')
            url.append(str(conf.get('extra')))
    elif conf.get('engine') == 'databricks':
        url.append(url_encode(str(conf.get('user'))))

        if conf.get('password'):
            url.append(':')
            url.append(url_encode(str(conf.get('password'))))

        url.append('@')

        url.append(str(conf.get('host')))

        if conf.get('port'):
            url.append(':')
            url.append(str(conf.get('port')))

        extra_dict = dict(parse_qsl(conf.get('extra'), keep_blank_values=True))

        url.append('?http_path={}'.format(url_encode(extra_dict.get('http_path'))))
        url.append('&catalog={}'.format(url_encode(extra_dict.get('catalog'))))
        url.append('&schema={}'.format(url_encode(conf.get('name'))))

        if 'http_path' in extra_dict:
            extra_dict.pop('http_path')

        if 'catalog' in extra_dict:
            extra_dict.pop('catalog')

        for key, value in extra_dict.items():
            url.append('&')
            url.append('&{}={}'.format(url_encode(key), url_encode(value)))
    else:
        host = '127.0.0.1' if tunnel else conf.get('host')
        port = tunnel.local_bind_port if tunnel else conf.get('port')

        if conf.get('user'):
            url.append(url_encode(str(conf.get('user'))))

            if conf.get('password'):
                url.append(':')
                url.append(url_encode(str(conf.get('password'))))

            if host:
                url.append('@')

        if host:
            url.append(str(host))

            if port:
                url.append(':')
                url.append(str(port))

            url.append('/')

        if conf.get('engine') != 'oracle':
            url.append(str(conf.get('name')))

        if conf.get('extra'):
            url.append('?')
            url.append(str(conf.get('extra')))
        elif conf.get('engine') == 'mysql':
            url.append('?charset=utf8')
        elif conf.get('engine') == 'mssql+pyodbc':
            url.append('?driver=FreeTDS')
        elif conf.get('engine') == 'oracle':
            url.append('?service_name={}'.format(url_encode(conf.get('name'))))
        elif conf.get('engine') == 'clickhouse+native':
            url.append('?secure=True')

    return ''.join(url)


def url_encode(value):
    return quote_plus(value)


def sql_create_connection_engine(conf, tunnel):
    engine_url = sql_build_engine_url(conf, tunnel)

    if not engine_url:
        raise Exception('Database configuration is not set')

    if conf.get('engine') == 'sqlite':
        return create_engine(engine_url)
    elif conf.get('engine') == 'mysql':
        connect_args = {}
        ssl = {
            'ca': conf.get('ssl_ca'),
            'cert': conf.get('ssl_cert'),
            'key': conf.get('ssl_key')
        }
        ssl_set = dict(list(filter(lambda x: x[1], ssl.items())))

        if len(ssl_set):
            connect_args['ssl'] = ssl_set

        return create_engine(
            engine_url,
            pool_size=conf.get('connections'),
            pool_pre_ping=True,
            max_overflow=conf.get('connections_overflow'),
            pool_recycle=300,
            connect_args={
                'connect_timeout': 5,
                **connect_args
            }
        )
    elif conf.get('engine') == 'bigquery':
        return create_engine(
            engine_url,
            pool_size=conf.get('connections'),
            pool_pre_ping=True,
            max_overflow=conf.get('connections_overflow'),
            pool_recycle=300
        )
    elif conf.get('engine') == 'oracle':
        return create_engine(
            engine_url,
            pool_size=conf.get('connections'),
            pool_pre_ping=True,
            max_overflow=conf.get('connections_overflow'),
            pool_recycle=300
        )
    else:
        return create_engine(
            engine_url,
            pool_size=conf.get('connections'),
            pool_pre_ping=True,
            max_overflow=conf.get('connections_overflow'),
            pool_recycle=300,
            connect_args={'connect_timeout': 5}
        )


def sql_load_mapped_base(MappedBase, clear=False):
    def classname_for_table(base, tablename, table):
        return get_table_name(MappedBase.metadata, table)

    def name_for_scalar_relationship(base, local_cls, referred_cls, constraint):
        foreign_key = constraint.elements[0] if len(constraint.elements) else None
        if foreign_key:
            name = '__'.join([foreign_key.parent.name, 'to', foreign_key.column.table.name, foreign_key.column.name])
        else:
            name = referred_cls.__name__.lower()

        if name in constraint.parent.columns:
            name = name + '_relation'
            logger.warning('Already detected column name, using {}'.format(name))

        return name

    def name_for_collection_relationship(base, local_cls, referred_cls, constraint):
        foreign_key = constraint.elements[0] if len(constraint.elements) else None
        if foreign_key:
            name = '__'.join([foreign_key.parent.table.name, foreign_key.parent.name, 'to', foreign_key.column.name])
        else:
            name = referred_cls.__name__.lower()

        if name in constraint.parent.columns:
            name = name + '_relation'
            logger.warning('Already detected column name, using {}'.format(name))

        return name

    if clear:
        MappedBase.registry.dispose()
        MappedBase.classes.clear()

    MappedBase.prepare(
        classname_for_table=classname_for_table,
        name_for_scalar_relationship=name_for_scalar_relationship,
        name_for_collection_relationship=name_for_collection_relationship
    )


def sql_load_database_table(engine, MappedBase, schema, table):
    bind = MappedBase.metadata.bind

    related_metadata = MetaData(schema=schema, bind=bind)
    related_metadata.reflect(bind=engine, schema=schema, only=[table])
    related_base = automap_base(metadata=related_metadata)
    sql_load_mapped_base(related_base)

    return related_base.classes.get(table)



from social_core.actions import do_auth

from jet_bridge_base.configuration import configuration
from jet_bridge_base.external_auth.mixin import ExternalAuthMixin
from jet_bridge_base.views.base.api import BaseAPIView

AUTH_URI_KEY = 'auth'
PROJECT_KEY = 'project'
REDIRECT_URI_KEY = 'redirect_uri'


class ExternalAuthLoginView(ExternalAuthMixin, BaseAPIView):

    def get(self, request, *args, **kwargs):
        app = kwargs.get('app')
        return self._auth(request, app)

    def post(self, request, *args, **kwargs):
        app = kwargs.get('app')
        return self._auth(request, app)

    def _auth(self, request, app):
        auth_uri = request.get_argument('auth_uri', None)
        project = request.get_argument('project', None)
        redirect_uri = request.get_argument('redirect_uri', None)

        configuration.session_set(request, AUTH_URI_KEY, auth_uri)
        configuration.session_set(request, PROJECT_KEY, project)
        configuration.session_set(request, REDIRECT_URI_KEY, redirect_uri)

        self.init_auth(request, app)
        return do_auth(self.backend)



from jet_bridge_base.exceptions.sql import SqlError
from jet_bridge_base.exceptions.validation_error import ValidationError
from jet_bridge_base.permissions import HasProjectPermissions
from jet_bridge_base.responses.json import JSONResponse
from jet_bridge_base.serializers.sql import SqlSerializer, SqlsSerializer
from jet_bridge_base.utils.track_database import track_database_async
from jet_bridge_base.views.base.api import APIView
from jet_bridge_base.status import HTTP_400_BAD_REQUEST


class SqlView(APIView):
    permission_classes = (HasProjectPermissions,)
    track_queries = True

    def post(self, request, *args, **kwargs):
        track_database_async(request)

        if 'queries' in request.data:
            serializer = SqlsSerializer(data=request.data, context={'request': request})
        else:
            serializer = SqlSerializer(data=request.data, context={'request': request})

        try:
            serializer.is_valid(raise_exception=True)
        except ValidationError as e:
            return JSONResponse({'error': str(e)}, status=HTTP_400_BAD_REQUEST)

        try:
            result = serializer.execute(serializer.validated_data)
            return JSONResponse(result)
        except SqlError as e:
            return JSONResponse({'error': str(e.detail)}, status=HTTP_400_BAD_REQUEST)
